#   Copyright 2024 Beatriz Lourenço, Miguel Loureiro, IS4
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
This module contains the Markov cross-validation method.

Classes
-------
MarkovCV
    Implements the Markov cross-validation method.
"""

from .base import BaseSplitter
from ..data_characteristics import get_features
import warnings
import numpy as np
import pandas as pd
from typing import Generator
import matplotlib.pyplot as plt
import math


class MarkovCV(BaseSplitter):
    """
    Implements the Markov cross-validation method.

    This class implements the Markov cross-validation method.

    Parameters
    ----------
    ts : np.ndarray | pd.Series
        Univariate time series.

    p : int
        p-order autocorrelation.

    seed : int, default=1
        Random seed.

    Attributes
    ----------
    n_splits : int
        The number of splits.

    sampling_freq : int | float
        The series' sampling frequency (Hz).

    Methods
    -------
    split()
        Split the time series into training and validation sets.

    info()            
        Provide additional information on the validation method.

    statistics() 
        Compute relevant statistics for both training and validation sets.

    plot(height: int, width: int)
        Plot the partitioned time series.

    Raises
    ------
    TypeError
        If `seed` is not an integer.

    TypeError
        If `p` is not an integer.

    ValueError
        If `p` is not positive.

    Notes
    -----
    The Markov cross-validation method partitions the data so that every partition can be regarded as \
    a Markov Process. It uses the linear autocorrelation measure to ensure that the samples in both the \
    training set and the validation set are neither too close nor too far apart.

    ![Markov](../../../images/Markov.png)

    For a thorough discussion of the method, see [[1]](#1).

    References
    ----------
    ##1
    Gaoxia Jiang and Wenjian Wang. Markov cross-validation for time series
    model evaluations. Information Sciences, 375:219–233, 2017
    """

    def __init__(self, ts: np.ndarray | pd.Series, p: int, seed: int = 1) -> None:
        self._check_seed(seed)
        self._check_p(p)

        if p % 3 == 0:
            self._m = math.floor(2 * p / 3) + 1
        else:
            self._m = math.floor(2 * p / 3) + 2

        self.n_subsets = (
            2 * self._m
        )  # total number of subsets (training + tests subsets)
        splits = 2 * self._m  # due to 2-fold CV
        super().__init__(splits, ts, 1)
        self._p = p
        self._seed = seed
        self._suo = {}
        self._sue = {}

    def _check_p(self, p: int) -> None:
        """
        Perform a type and value check on the p.
        """

        if isinstance(p, int) is False:

            raise TypeError("'p' should be an integer.")

        if p < 0:

            raise ValueError("'p' must be positive.")

        return

    def _check_seed(self, seed: int) -> None:
        """
        Perform a type check on the seed.
        """

        if isinstance(seed, int) is False:

            raise TypeError("'seed' should be an integer.")

        return

    @property
    def sampling_freq(self) -> int | float:
        """
        Get the time series' sampling frequency.

        This method can be used to access the time series' sampling
        frequency, in Hertz (this is set on intialisation).
        Since the method is implemented as a property, this information can
        simply be accessed as an attribute using dot notation.

        Returns
        -------
        int | float
            The time series' sampling frequency (Hz).
        """
        warnings.warn(
            "The training and tests sets generated by this method do not maintain a constant sampling rate.",
            UserWarning,
        )

        return None

    def _markov_iteration(self, n: int) -> np.array:
        np.random.seed(self._seed)

        d = np.zeros(n, dtype=int)

        i, j = 1, -1
        if np.random.rand() < 0.25:
            d[0], d[1] = i, i + 1
        elif np.random.rand() < 0.5:
            d[0], d[1] = i, j - 1
        elif np.random.rand() < 0.75:
            d[0], d[1] = j - 1, i
        else:
            d[0], d[1] = j - 1, j - 1

        for t in range(2, n):
            rd = np.random.rand()
            if (d[t - 1] > 0) and (d[t - 2] > 0):
                d[t] = j
                j -= 1
            elif (d[t - 1] < 0) and (d[t - 2] < 0):
                d[t] = i
                i += 1
            elif rd > 0.5:
                d[t] = j
                j -= 1
            else:
                d[t] = i
                i += 1

        return d

    def _markov_partitions(self):
        n = self._n_samples

        d = self._markov_iteration(n)

        Id = np.mod(d, self._m) + 1 + np.where(d > 0, 1, 0) * self._m

        Su = {}
        for u in range(1, 2 * self._m + 1):
            Su[u] = np.where(Id == u)[0]

        self._suo = {}
        self._sue = {}
        for u in range(1, self._m + 1):
            self._suo[u] = Su[u * 2 - 1]
            self._sue[u] = Su[u * 2]

    def split(self) -> Generator[tuple[np.ndarray, np.ndarray, float], None, None]:
        """
        Split the time series into training and validation sets.

        This method splits the series' indices into disjoint sets containing the training and validation indices.
        At every iteration, an array of training indices and another one containing the validation indices are generated.
        Note that this method is a generator. To access the indices, use the `next()` method or a `for` loop.

        Yields
        ------
        np.ndarray
            Array of training indices.

        np.ndarray
            Array of validation indices.

        float
            Used for compatibility reasons. Irrelevant for this method.

        Examples
        --------
        >>> import numpy as np
        >>> from timecave.validation_methods.markov import MarkovCV
        >>> ts = np.ones(10);
        >>> splitter = MarkovCV(ts, p=2);
        >>> for ind, (train, val, _) in enumerate(splitter.split()):
        ... 
        ...     print(f"Iteration {ind+1}");
        ...     print(f"Training set indices: {train}");
        ...     print(f"Validation set indices: {val}");
        Iteration 1
        Training set indices: [8]
        Validation set indices: [0 6]
        Iteration 2
        Training set indices: [0 6]
        Validation set indices: [8]
        Iteration 3
        Training set indices: [3]
        Validation set indices: [5]
        Iteration 4
        Training set indices: [5]
        Validation set indices: [3]
        Iteration 5
        Training set indices: [1 2 7]
        Validation set indices: [4 9]
        Iteration 6
        Training set indices: [4 9]
        Validation set indices: [1 2 7]
        """

        self._markov_partitions()
        for i in range(1, len(self._suo.items()) + 1):
            train, validation = self._suo[i], self._sue[i]
            yield (train, validation, 1.0)
            train, validation = self._sue[i], self._suo[i]
            yield (train, validation, 1.0)  # two-fold cross validation

    def info(self) -> None:
        """
        Provide some basic information on the training and validation sets.

        This method displays the number of splits and the number of observations per set.

        Examples
        --------
        >>> import numpy as np
        >>> from timecave.validation_methods.markov import MarkovCV
        >>> ts = np.ones(10);
        >>> splitter = MarkovCV(ts, p=2);
        >>> splitter.info();
        Markov CV method
        ---------------
        Time series size: 10 samples
        Number of splits: 6
        Number of observations per set: 1 to 3
        """

        self._markov_partitions()

        lengths = []
        for i in range(1, len(self._suo.items()) + 1):
            lengths.extend([len(self._suo[i]), len(self._sue[i])])

        print("Markov CV method")
        print("---------------")
        print(f"Time series size: {self._n_samples} samples")
        print(f"Number of splits: {self.n_splits}")
        print(f"Number of observations per set: {min(lengths)} to {max(lengths)}")
        pass

    def statistics(self) -> tuple[pd.DataFrame]:
        """
        Compute relevant statistics for both training and validation sets.

        This method computes relevant time series features, such as mean, strength-of-trend, etc. for both the whole time series, the training set and the validation set.
        It can and should be used to ensure that the characteristics of both the training and validation sets are, statistically speaking, similar to those of the time series one wishes to forecast.
        If this is not the case, using the validation method will most likely lead to a poor assessment of the model's performance.

        Returns
        -------
        pd.DataFrame
            Relevant features for the entire time series.

        pd.DataFrame
            Relevant features for the training set.

        pd.DataFrame
            Relevant features for the validation set.

        Raises
        ------
        ValueError
            If the time series is composed of less than three samples.
        
        ValueError
            If the folds comprise less than two samples.

        Examples
        --------

        Frequency-domain features are not computed for the Markov CV method:

        >>> import numpy as np
        >>> from timecave.validation_methods.markov import MarkovCV
        >>> ts = np.hstack((np.ones(5), np.zeros(5)));
        >>> splitter = MarkovCV(ts, p=1);
        >>> ts_stats, training_stats, validation_stats = splitter.statistics();
        Frequency features are only meaningful if the correct sampling frequency is passed to the class.
        >>> ts_stats
           Mean  Median  Min  Max  Variance  P2P_amplitude  Trend_slope  Strength_of_trend  Mean_crossing_rate  Median_crossing_rate
        0   0.5     0.5  0.0  1.0      0.25            1.0    -0.151515            1.59099            0.111111              0.111111
        >>> training_stats
           Mean  Median  Min  Max  Variance  P2P_amplitude  Trend_slope  Strength_of_trend  Mean_crossing_rate  Median_crossing_rate
        0   0.5     0.5  0.0  1.0      0.25            1.0         -1.0                inf            1.000000              1.000000
        0   0.5     0.5  0.0  1.0      0.25            1.0         -1.0                inf            1.000000              1.000000
        0   0.5     0.5  0.0  1.0      0.25            1.0         -1.0                inf            1.000000              1.000000
        0   0.5     0.5  0.0  1.0      0.25            1.0         -0.4            1.06066            0.333333              0.333333
        >>> validation_stats
           Mean  Median  Min  Max  Variance  P2P_amplitude  Trend_slope  Strength_of_trend  Mean_crossing_rate  Median_crossing_rate
        0   0.5     0.5  0.0  1.0      0.25            1.0         -1.0                inf            1.000000              1.000000
        0   0.5     0.5  0.0  1.0      0.25            1.0         -1.0                inf            1.000000              1.000000
        0   0.5     0.5  0.0  1.0      0.25            1.0         -0.4            1.06066            0.333333              0.333333
        0   0.5     0.5  0.0  1.0      0.25            1.0         -1.0                inf            1.000000              1.000000
        """

        columns = [
            "Mean",
            "Median",
            "Min",
            "Max",
            "Variance",
            "P2P_amplitude",
            "Trend_slope",
            "Strength_of_trend",
            "Mean_crossing_rate",
            "Median_crossing_rate",
        ]

        if self._n_samples <= 2:

            raise ValueError(
                "Basic statistics can only be computed if the time series comprises more than two samples."
            )

        print("Frequency features are only meaningful if the correct sampling frequency is passed to the class.")

        full_features = get_features(self._series, self._fs)[columns]
        training_stats = []
        validation_stats = []

        for training, validation, _ in self.split():

            training_feat = get_features(self._series[training], self._fs)
            training_stats.append(training_feat[columns])

            validation_feat = get_features(self._series[validation], self._fs)
            validation_stats.append(validation_feat[columns])

        training_features = pd.concat(training_stats)
        validation_features = pd.concat(validation_stats)

        return (full_features, training_features, validation_features)

    def plot(self, height: int, width: int) -> None:
        """
        Plot the partitioned time series.

        This method allows the user to plot the partitioned time series. The training and validation sets are plotted using different colours.

        Parameters
        ----------
        height : int
            The figure's height.

        width : int
            The figure's width.

        Examples
        --------
        >>> import numpy as np
        >>> from timecave.validation_methods.markov import MarkovCV
        >>> ts = np.ones(100);
        >>> splitter = MarkovCV(ts, p=1);
        >>> splitter.plot(10, 10);

        ![markov_plot](../../../images/Markov_plot.png)
        """

        fig, axs = plt.subplots(self.n_splits, 1, sharex=True)
        fig.set_figheight(height)
        fig.set_figwidth(width)
        fig.supxlabel("Samples")
        fig.supylabel("Time Series")
        fig.suptitle("Markov CV method")

        for it, (training, validation, _) in enumerate(self.split()):

            axs[it].scatter(training, self._series[training], label="Training set")
            axs[it].scatter(
                validation, self._series[validation], label="Validation set"
            )
            axs[it].set_title("Iteration {}".format(it + 1))
            axs[it].legend()

        plt.subplots_adjust(hspace=0.5)
        plt.show()

        return


if __name__ == "__main__":
    #mcv = MarkovCV(ts=np.arange(50), p=1, seed=1)
    #mcv.split()
    #mcv.info()
    #mcv.plot(2, 10)
    #print()

    import doctest

    doctest.testmod(verbose=True);